from __future__ import annotations

import time

from typing import List

from bs4 import BeautifulSoup

from instascrape.core._mappings import _PostMapping, _ProfileMapping
from instascrape.core._static_scraper import _StaticHtmlScraper
from instascrape.scrapers.post import Post

class Profile(_StaticHtmlScraper):
    """Scraper for an Instagram profile page"""

    _Mapping = _ProfileMapping

    def get_recent_posts(self, amt: int = 12) -> List[Post]:
        """
        Return a list of the profiles recent posts. Max available for return
        is 12.

        Parameters
        ----------
        amt : int
            Amount of recent posts to return

        Returns
        -------
        posts : List[Post]
            List containing the recent 12 posts and their available data
        """
        if amt > 12:
            raise IndexError(
                f"{amt} is too large, 12 is max available posts. Getting more posts will require an out-of-the-box extension."
            )
        posts = []
        try:
            post_arr = self.json_dict["entry_data"]["ProfilePage"][0]["graphql"]["user"][
                "edge_owner_to_timeline_media"
            ]["edges"]
        except TypeError:
            raise ValueError(
                "Can't return posts without first scraping the Profile. Call the scrape method on your object first."
            )

        for post in post_arr[:amt]:
            json_dict = post["node"]
            mapping = _PostMapping.post_from_profile_mapping()
            post = Post(json_dict)
            post.scrape(mapping=mapping)
            post.username = self.username
            post.full_name = self.full_name
            posts.append(post)
        return posts

    def get_posts(self, webdriver, amount=None, login_first=False, login_pause=60, max_failed_scroll=300, scrape=False, scrape_pause=5):
        """
        Return Post objects from profile scraped using a webdriver (not included)

        Parameters
        ----------
        webdriver : selenium.webdriver.chrome.webdriver.WebDriver
            Selenium webdriver for rendering JavaScript and loading dynamic
            content
        amount : int
            Amount of posts to return, default is all of them
        login_first : bool
            Start on login page to allow user to manually login to Instagram
        login_pause : int
            Length of time in seconds to pause before starting scrape
        max_failed_scroll : int
            Maximum amount of scroll attempts before stopping if scroll is stuck
        scrape : bool
            Scrape posts with the webdriver prior to returning
        scrape_pause : int
            Time in seconds between each scrape

        Returns
        -------
        posts : List[Post]
            Post objects gathered from the profile page
        """

        JS_SCROLL_SCRIPT = "window.scrollTo(0, document.body.scrollHeight); var lenOfPage=document.body.scrollHeight; return lenOfPage;"
        JS_PAGE_LENGTH_SCRIPT = "var lenOfPage=document.body.scrollHeight; return lenOfPage;"

        # Determine how many posts are available on the page
        try:
            posts_len = self.posts
        except AttributeError:
            raise AttributeError(f"{type(self)} must be scraped first")

        # Manual login
        if login_first:
            webdriver.get("https://www.instagram.com")
            time.sleep(login_pause)

        # Get profile page
        webdriver.get(self.url)

        # Continuously scroll, collect HTML, and parse Post objects
        posts = []
        shortcodes = []
        scroll_attempts = 0
        last_position = webdriver.execute_script(JS_PAGE_LENGTH_SCRIPT)
        scrolling = True
        while scrolling:
            current_position = webdriver.execute_script(JS_SCROLL_SCRIPT)
            source_data = webdriver.page_source
            found_posts = self._separate_posts(source_data)

            # Append found posts into total posts
            for post in found_posts:
                if post.source not in shortcodes:
                    shortcodes.append(post.source)
                    posts.append(post)

            # If scroll is stuck and exceeds max allowed attempts, exit loop
            if current_position == last_position:
                scroll_attempts += 1
                if scroll_attempts > max_failed_scroll:
                    scrolling = False
            else:
                scroll_attempts = 0
                last_position = current_position

            current_post_len = len(posts)
            if (current_post_len >= posts_len) or (current_post_len >= amount):
                break

        # Remove excess posts from right of list
        posts = posts[:amount]

        # If scrape arg is True, scrape all posts using webdriver
        scraped_posts = []
        if scrape:
            for post in posts:
                scraped_posts.append(post.scrape(inplace=False, webdriver=webdriver))
                time.sleep(scrape_pause)
            posts = scraped_posts

        return posts

    def _separate_posts(self, source_data):
        """Separate the HTML and parse out BeautifulSoup for every post"""
        post_soup = []

        soup = BeautifulSoup(source_data, features="lxml")
        anchor_tags = soup.find_all("a")
        post_tags = [tag for tag in anchor_tags if tag.find(
            "div", {"class": "eLAPa"})]

        #Filter new posts that have not been stored yet
        new_posts = [tag for tag in post_tags if tag not in post_soup]
        post_soup += new_posts

        return self._create_post_objects(post_soup)

    def _create_post_objects(self, post_soup):
        """Create a Post object from the given shortcode"""
        posts = []
        for post in post_soup:
            shortcode = post["href"].replace("/p/", "")[:-1]
            posts.append(Post(shortcode))
        return posts

    def _url_from_suburl(self, suburl):
        return f"https://www.instagram.com/{suburl}/"



"""
Mappings that tell the _JsonEngine the user facing attribute names and the
steps needed to get there in a JSON dictionary
"""

from __future__ import annotations

from abc import ABC
from collections import deque
from copy import deepcopy
from typing import Dict, List, Union

# pylint: disable=used-before-assignment


MappingObject = Union["_PostMapping", "_ProfileMapping", "_HashtagMapping", "_LoginMapping"]


class _GeneralMapping(ABC):
    """
    Maps the user interfacing attribute names with their keys as given in a JSON
    dict that has been flattened using
    instascrape.core._json_flattener.JsonFlattener

    Attributes
    ----------
    mapping : Dict[str, deque]
        Each key: val pair represents one data point and the directive for
        traversing a JSON dict and accessing that value

    Methods
    -------
    return_mapping(keys: List[str]=[]) -> Dict[str, deque]
        Interface for returning only mapping directives that are specified in
        a list of keys

    """

    mapping = {
        # "csrf_token": deque(["csrf_token"]),
        # "viewer_id": deque(["viewerId"]),
        # "country_code": deque(["country_code"]),
        # "language_code": deque(["language_code"]),
        # "locale": deque(["locale"]),
        # "device_id": deque(["device_id"]),
        # "browser_push_pub_key": deque(["browser_push_pub_key"]),
        # "key_id": deque(["key_id"]),
        # "public_key": deque(["public_key"]),
        # "version": deque(["version"]),
        # "is_dev": deque(["is_dev"]),
        # "rollout_hash": deque(["rollout_hash"]),
        # "bundle_variant": deque(["bundle_variant"]),
        # "frontend_dev": deque(["frontend_env"]),
    }

    @classmethod
    def return_mapping(cls, keys: List[str] = None, exclude: List[str] = None) -> Dict[str, deque]:
        """
        Return key-directive pairs specified by key names. If no keys are
        specified, return all

        Parameters
        ----------
        keys : List[str]
            Keys that specify what directives to return

        Returns
        -------
        directive_dict : Dict[str, deque]
            Dictionary of keys and their directives
        """
        if keys is None:
            keys = []
        if exclude is None:
            exclude = []
        if isinstance(keys, str):
            keys = [keys]
        if isinstance(exclude, str):
            exclude = [exclude]

        if not keys:
            keys = list(cls.mapping)
        if exclude:
            keys = [key for key in keys if key not in exclude]
        directive_dict = {key: deepcopy(cls.mapping[key]) for key in keys}
        return directive_dict


class _PostMapping(_GeneralMapping):
    """Mapping specific to Instagram post pages"""

    mapping = _GeneralMapping.return_mapping().copy()
    mapping.update(
        {
            "id": deque(["id"]),
            "shortcode": deque(["shortcode"]),
            "height": deque(["height"]),
            "width": deque(["width"]),
            "gating_info": deque(["gating_info"]),
            "fact_check_overall_rating": deque(["fact_check_overall_rating"]),
            "fact_check_information": deque(["fact_check_information"]),
            "sensitivity_friction_info": deque(["sensitivity_friction_info"]),
            "media_overlay_info": deque(["media_overlay_info"]),
            "media_preview": deque(["media_preview"]),
            "display_url": deque(["display_url"]),
            "accessibility_caption": deque(["accessibility_caption"]),
            "is_video": deque(["is_video"]),
            "tracking_token": deque(["tracking_token"]),
            "tagged_users": deque(["edge_media_to_tagged_user"]),
            "caption": deque(["text"]),
            "caption_is_edited": deque(["caption_is_edited"]),
            "has_ranked_comments": deque(["has_ranked_comments"]),
            "comments": deque(["count"]),
            "comments_disabled": deque(["comments_disabled"]),
            "commenting_disabled_for_viewer": deque(["commenting_disabled_for_viewer"]),
            "timestamp": deque(["taken_at_timestamp"]),
            "likes": deque(["edge_media_preview_like_count"]),
            "location": deque(["name"]),
            "viewer_has_liked": deque(["viewer_has_liked"]),
            "viewer_has_saved": deque(["viewer_has_saved"]),
            "viewer_has_saved_to_collection": deque(["viewer_has_saved_to_collection"]),
            "viewer_in_photo_of_you": deque(["viewer_in_photo_of_you"]),
            "viewer_can_reshare": deque(["viewer_can_reshare"]),
            "video_url": deque(["video_url"]),
            "has_audio": deque(["has_audio"]),
            "video_view_count": deque(["video_view_count"]),
            "username": deque(["shortcode_media_owner_username"]),
            "full_name": deque(['owner_full_name']),
        }
    )

    @classmethod
    def post_from_profile_mapping(cls):
        """
        Return the mapping needed for parsing a post's JSON data from the JSON
        served back after requesting a Profile page.
        """
        return {
            "id": deque(["id"]),
            "shortcode": deque(["shortcode"]),
            "dimensions": deque(["dimensions"]),
            "display_url": deque(["display_url"]),
            "tagged_users": deque(["edge_media_to_tagged_user", "edges"]),
            "fact_check_overall_rating": deque(["fact_check_overall_rating"]),
            "fact_check_information": deque(["fact_check_information"]),
            "is_video": deque(["is_video"]),
            "accessibility_caption": deque(["accessibility_caption"]),
            "caption": deque(["edge_media_to_caption", "edges", 0, "node", "text"]),
            "comments": deque(["edge_media_to_comment", "count"]),
            "comments_disabled": deque(["comments_disabled"]),
            "upload_date": deque(["taken_at_timestamp"]),
            "likes": deque(["edge_media_preview_like", "count"]),
            "location": deque(["location"]),
        }

    @classmethod
    def post_from_hashtag_mapping(cls):
        """
        Return the mapping needed for parsing a post's JSON data from the JSON
        served back after requesting a Hashtag page.
        """
        return {
            "comments_disabled": deque(["comments_disabled"]),
            "id": deque(["id"]),
            "caption": deque(["edge_media_to_caption", "edges", 0, "node", "text"]),
            "shortcode": deque(["shortcode"]),
            "comments": deque(["edge_media_to_comment", "count"]),
            "upload_date": deque(["taken_at_timestamp"]),
            "dimensions": deque(["dimensions"]),
            "display_url": deque(["display_url"]),
            "likes": deque(["edge_media_preview_like", "count"]),
            "owner": deque(["owner", "id"]),
            "is_video": deque(["is_video"]),
            "accessibility_caption": deque(["accessibility_caption"]),
        }


class _ReelMapping(_PostMapping):
    mapping = _PostMapping.return_mapping().copy()
    mapping.update(
        {
            "video_play_count": deque(["video_play_count"]),
        }
    )


class _IGTVMapping(_PostMapping):
    mapping = _PostMapping.return_mapping().copy()


class _ProfileMapping(_GeneralMapping):
    """Mapping specific to Instagram profile pages"""

    mapping = _GeneralMapping.return_mapping().copy()
    mapping.update(
        {
            "logging_page_id": deque(["logging_page_id"]),
            "show_suggested_profiles": deque(["show_suggested_profiles"]),
            "show_follow_dialog": deque(["show_follow_dialog"]),
            "biography": deque(["biography"]),
            "blocked_by_viewer": deque(["blocked_by_viewer"]),
            "restricted_by_viewer": deque(["restricted_by_viewer"]),
            "country_block": deque(["country_block"]),
            "external_url": deque(["external_url"]),
            "external_url_linkshimmed": deque(["external_url_linkshimmed"]),
            "followers": deque(["count"]),
            "followed_by_viewer": deque(["followed_by_viewer"]),
            "following": deque(["edge_follow_count"]),
            "follows_viewer": deque(["follows_viewer"]),
            "full_name": deque(["user_full_name"]),
            "has_ar_effects": deque(["has_ar_effects"]),
            "has_clips": deque(["has_clips"]),
            "has_guides": deque(["has_guides"]),
            "has_channel": deque(["has_channel"]),
            "has_blocked_viewer": deque(["has_blocked_viewer"]),
            "highlight_reel_count": deque(["highlight_reel_count"]),
            "has_requested_viewer": deque(["has_requested_viewer"]),
            "id": deque(["id"]),
            "is_business_account": deque(["is_business_account"]),
            "is_joined_recently": deque(["is_joined_recently"]),
            "business_category_name": deque(["business_category_name"]),
            "overall_category_name": deque(["overall_category_name"]),
            "category_enum": deque(["category_enum"]),
            "is_private": deque(["is_private"]),
            "is_verified": deque(["is_verified"]),
            "mutual_followers": deque(["edge_mutual_followed_by_count"]),
            "profile_pic_url": deque(["profile_pic_url"]),
            "profile_pic_url_hd": deque(["profile_pic_url_hd"]),
            "requested_by_viewer": deque(["requested_by_viewer"]),
            "username": deque(["user_username"]),
            "connected_fb_page": deque(["connected_fb_page"]),
            "posts": deque(["edge_owner_to_timeline_media_count"]),
        }
    )


class _HashtagMapping(_GeneralMapping):
    """Mapping specific to Instagram hashtag pages"""

    mapping = _GeneralMapping.return_mapping().copy()
    mapping.update(
        {
            "id": deque(["id"]),
            "name": deque(["name"]),
            "allow_following": deque(["allow_following"]),
            "is_following": deque(["is_following"]),
            "is_top_media_only": deque(["is_top_media_only"]),
            "profile_pic_url": deque(["profile_pic_url"]),
            "amount_of_posts": deque(["count"]),
        }
    )


class _LocationMapping(_GeneralMapping):
    """Mapping specific to Instagram profile pages"""

    mapping = _GeneralMapping.return_mapping().copy()
    mapping.update(
        {
            "id": deque(["id"]),
            "name": deque(["name"]),
            "has_public_page": deque(["has_public_page"]),
            "latitude": deque(["lat"]),
            "longitude": deque(["lng"]),
            "slug": deque(["slug"]),
            "blurb": deque(["blurb"]),
            "website": deque(["website"]),
            "phone": deque(["phone"]),
            "primary_alias_on_fb": deque(["primary_alias_on_fb"]),
            "stress_address": deque(["street_address"]),
            "zip_code": deque(["zip_code"]),
            "city_name": deque(["city_name"]),
            "region_name": deque(["region_name"]),
            "country_code": deque(["country_code"]),
            "amount_of_posts": deque(["count"]),
        }
    )


class _LoginMapping(_GeneralMapping):
    """Mapping specific to Instagram login page"""

    mapping = _GeneralMapping.return_mapping().copy()


class _HttpErrorMapping(_GeneralMapping):
    """Mapping specific to Instagram login page"""

    mapping = _GeneralMapping.return_mapping().copy()


class _MetaMapping:
    """
    Map the string in the Instagram JSON that indicates the type of page the
    JSON was scraped from

    Attributes
    ----------
    str_to_mapper_obj : Dict[str, Any]
        Dictionary that maps the string name of the JSON type to the specific
        mapping object

    Methods
    -------
    get_mapper(page_type: str)
        Return the mapping object that correlates to the string
    """

    str_to_mapper_obj = {
        "ProfilePage": _ProfileMapping,
        "TagPage": _HashtagMapping,
        "PostPage": _PostMapping,
        "LoginAndSignupPage": _LoginMapping,
        "LocationsPage": _LocationMapping
    }

    @classmethod
    def get_mapper(cls, page_type: str) -> MappingObject:
        """
        Return the appropriate mapper that corresponds to the page_type as
        given in the requested Instagram JSON data
        """
        return cls.str_to_mapper_obj[page_type]



"""
Post
----
    Scrape data from a Post page
"""
from __future__ import annotations

import datetime
from typing import List
import re
import shutil
import pathlib
import math

import requests

from instascrape.core._mappings import _PostMapping
from instascrape.core._static_scraper import _StaticHtmlScraper
from instascrape.scrapers.scrape_tools import parse_data_from_json
from instascrape.scrapers.comment import Comment

class Post(_StaticHtmlScraper):
    """Scraper for an Instagram post page"""

    _Mapping = _PostMapping
    SUPPORTED_DOWNLOAD_EXTENSIONS = [".mp3", ".mp4", ".png", ".jpg"]

    def scrape(
            self,
            mapping=None,
            keys: List[str] = None,
            exclude: List[str] = None,
            headers={
                "user-agent": "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Mobile Safari/537.36 Edg/87.0.664.57"
            },
            inplace=True,
            session=None,
            webdriver=None
        ) -> None:
        """
        Scrape data from the source

        Parameters
        ----------
        mapping : Dict[str, deque]
            Dictionary of parsing queue's that tell the JSON engine how to
            process the JSON data
        keys : List[str]
            List of strings that correspond to desired attributes for scraping
        exclude : List[str]
            List of strings that correspond to which attributes to exclude from
            being scraped
        headers : Dict[str, str]
            Dictionary of request headers to be passed on the GET request
        inplace : bool
            Determines if data modified inplace or return a new object with the
            scraped data
        session : requests.Session
            Session for making the GET request
        webdriver : selenium.webdriver.chrome.webdriver.WebDriver
            Webdriver for scraping the page, overrides any default or passed
            session

        Returns
        -------
        return_instance
            Optionally returns a scraped instance instead of modifying inplace
            if inplace arg is True
        """
        # pylint: disable=no-member, attribute-defined-outside-init
        if hasattr(self, "shortcode"):
            self.source = self.shortcode
        return_instance = super().scrape(
                            mapping=mapping,
                            keys=keys,
                            exclude=exclude,
                            headers=headers,
                            inplace=inplace,
                            session=session,
                            webdriver=webdriver
                        )
        if return_instance is None:
            return_instance = self

        # HACK: This isn't a very clean solution and there is certainly a better
        # way to deal with returning a Post object with only partial data
        if hasattr(return_instance, "timestamp"):
            return_instance.upload_date = datetime.datetime.fromtimestamp(return_instance.timestamp)
        if hasattr(return_instance, "shortcode"):
            return_instance.url = self._url_from_suburl(return_instance.shortcode)

        if mapping is None:
            return_instance.tagged_users = return_instance._parse_tagged_users(return_instance.json_dict)
            return_instance.hashtags = return_instance._parse_hashtags(return_instance.caption) if isinstance(return_instance.caption, str) else float("nan")
            try:
                if math.isnan(return_instance.full_name):
                    return_instance.full_name = return_instance.flat_json_dict["full_name"]
            except TypeError:
                pass
        return return_instance if return_instance is not self else None

    def download(self, fp: str) -> None:
        """
        Download an image or video from a post to your local machine at the given filepath

        Parameters
        ----------
        fp : str
            Filepath to download the image to
        """
        # pylint: disable=no-member

        ext = pathlib.Path(fp).suffix
        if ext not in self.SUPPORTED_DOWNLOAD_EXTENSIONS:
            raise NameError(
                f"{ext} is not a supported file extension. Please use {', '.join(self.SUPPORTED_DOWNLOAD_EXTENSIONS)}"
            )
        url = self.video_url if self.is_video else self.display_url

        resp = requests.get(url, stream=True)
        if not self.is_video:
            self._download_photo(fp, resp)
        else:
            self._download_video(fp, resp)

    def get_recent_comments(self) -> List[Comment]:
        """
        Returns a list of Comment objects that contain data regarding
        some of the posts comments

        Returns
        -------
        comments_arr : List[Comment]
            List of Comment objects
        """
        list_of_dicts = self.json_dict["entry_data"]["PostPage"][0]["graphql"]["shortcode_media"][
            "edge_media_to_parent_comment"
        ]["edges"]
        comments_arr = [Comment(comment_dict) for comment_dict in list_of_dicts]
        return comments_arr

    def embed(self) -> str:
        """
        Return embeddable HTML str for this post

        Returns
        -------
        html_template : str
            HTML string with embed markup for this Post
        """
        html_template = f'<blockquote class="instagram-media" data-instgrm-captioned data-instgrm-permalink="{self.url}?utm_source=ig_embed&amp;utm_campaign=loading" data-instgrm-version="13" style=" background:#FFF; border:0; border-radius:3px; box-shadow:0 0 1px 0 rgba(0,0,0,0.5),0 1px 10px 0 rgba(0,0,0,0.15); margin: 1px; max-width:540px; min-width:326px; padding:0; width:99.375%; width:-webkit-calc(100% - 2px); width:calc(100% - 2px);"><div style="padding:16px;"> <a href="{self.url}?utm_source=ig_embed&amp;utm_campaign=loading" style=" background:#FFFFFF; line-height:0; padding:0 0; text-align:center; text-decoration:none; width:100%;" target="_blank"> <div style=" display: flex; flex-direction: row; align-items: center;"> <div style="background-color: #F4F4F4; border-radius: 50%; flex-grow: 0; height: 40px; margin-right: 14px; width: 40px;"></div> <div style="display: flex; flex-direction: column; flex-grow: 1; justify-content: center;"> <div style=" background-color: #F4F4F4; border-radius: 4px; flex-grow: 0; height: 14px; margin-bottom: 6px; width: 100px;"></div> <div style=" background-color: #F4F4F4; border-radius: 4px; flex-grow: 0; height: 14px; width: 60px;"></div></div></div><div style="padding: 19% 0;"></div> <div style="display:block; height:50px; margin:0 auto 12px; width:50px;"><svg width="50px" height="50px" viewBox="0 0 60 60" version="1.1" xmlns="https://www.w3.org/2000/svg" xmlns:xlink="https://www.w3.org/1999/xlink"><g stroke="none" stroke-width="1" fill="none" fill-rule="evenodd"><g transform="translate(-511.000000, -20.000000)" fill="#000000"><g><path d="M556.869,30.41 C554.814,30.41 553.148,32.076 553.148,34.131 C553.148,36.186 554.814,37.852 556.869,37.852 C558.924,37.852 560.59,36.186 560.59,34.131 C560.59,32.076 558.924,30.41 556.869,30.41 M541,60.657 C535.114,60.657 530.342,55.887 530.342,50 C530.342,44.114 535.114,39.342 541,39.342 C546.887,39.342 551.658,44.114 551.658,50 C551.658,55.887 546.887,60.657 541,60.657 M541,33.886 C532.1,33.886 524.886,41.1 524.886,50 C524.886,58.899 532.1,66.113 541,66.113 C549.9,66.113 557.115,58.899 557.115,50 C557.115,41.1 549.9,33.886 541,33.886 M565.378,62.101 C565.244,65.022 564.756,66.606 564.346,67.663 C563.803,69.06 563.154,70.057 562.106,71.106 C561.058,72.155 560.06,72.803 558.662,73.347 C557.607,73.757 556.021,74.244 553.102,74.378 C549.944,74.521 548.997,74.552 541,74.552 C533.003,74.552 532.056,74.521 528.898,74.378 C525.979,74.244 524.393,73.757 523.338,73.347 C521.94,72.803 520.942,72.155 519.894,71.106 C518.846,70.057 518.197,69.06 517.654,67.663 C517.244,66.606 516.755,65.022 516.623,62.101 C516.479,58.943 516.448,57.996 516.448,50 C516.448,42.003 516.479,41.056 516.623,37.899 C516.755,34.978 517.244,33.391 517.654,32.338 C518.197,30.938 518.846,29.942 519.894,28.894 C520.942,27.846 521.94,27.196 523.338,26.654 C524.393,26.244 525.979,25.756 528.898,25.623 C532.057,25.479 533.004,25.448 541,25.448 C548.997,25.448 549.943,25.479 553.102,25.623 C556.021,25.756 557.607,26.244 558.662,26.654 C560.06,27.196 561.058,27.846 562.106,28.894 C563.154,29.942 563.803,30.938 564.346,32.338 C564.756,33.391 565.244,34.978 565.378,37.899 C565.522,41.056 565.552,42.003 565.552,50 C565.552,57.996 565.522,58.943 565.378,62.101 M570.82,37.631 C570.674,34.438 570.167,32.258 569.425,30.349 C568.659,28.377 567.633,26.702 565.965,25.035 C564.297,23.368 562.623,22.342 560.652,21.575 C558.743,20.834 556.562,20.326 553.369,20.18 C550.169,20.033 549.148,20 541,20 C532.853,20 531.831,20.033 528.631,20.18 C525.438,20.326 523.257,20.834 521.349,21.575 C519.376,22.342 517.703,23.368 516.035,25.035 C514.368,26.702 513.342,28.377 512.574,30.349 C511.834,32.258 511.326,34.438 511.181,37.631 C511.035,40.831 511,41.851 511,50 C511,58.147 511.035,59.17 511.181,62.369 C511.326,65.562 511.834,67.743 512.574,69.651 C513.342,71.625 514.368,73.296 516.035,74.965 C517.703,76.634 519.376,77.658 521.349,78.425 C523.257,79.167 525.438,79.673 528.631,79.82 C531.831,79.965 532.853,80.001 541,80.001 C549.148,80.001 550.169,79.965 553.369,79.82 C556.562,79.673 558.743,79.167 560.652,78.425 C562.623,77.658 564.297,76.634 565.965,74.965 C567.633,73.296 568.659,71.625 569.425,69.651 C570.167,67.743 570.674,65.562 570.82,62.369 C570.966,59.17 571,58.147 571,50 C571,41.851 570.966,40.831 570.82,37.631"></path></g></g></g></svg></div><div style="padding-top: 8px;"> <div style=" color:#3897f0; font-family:Arial,sans-serif; font-size:14px; font-style:normal; font-weight:550; line-height:18px;"> View this post on Instagram</div></div><div style="padding: 12.5% 0;"></div> <div style="display: flex; flex-direction: row; margin-bottom: 14px; align-items: center;"><div> <div style="background-color: #F4F4F4; border-radius: 50%; height: 12.5px; width: 12.5px; transform: translateX(0px) translateY(7px);"></div> <div style="background-color: #F4F4F4; height: 12.5px; transform: rotate(-45deg) translateX(3px) translateY(1px); width: 12.5px; flex-grow: 0; margin-right: 14px; margin-left: 2px;"></div> <div style="background-color: #F4F4F4; border-radius: 50%; height: 12.5px; width: 12.5px; transform: translateX(9px) translateY(-18px);"></div></div><div style="margin-left: 8px;"> <div style=" background-color: #F4F4F4; border-radius: 50%; flex-grow: 0; height: 20px; width: 20px;"></div> <div style=" width: 0; height: 0; border-top: 2px solid transparent; border-left: 6px solid #f4f4f4; border-bottom: 2px solid transparent; transform: translateX(16px) translateY(-4px) rotate(30deg)"></div></div><div style="margin-left: auto;"> <div style=" width: 0px; border-top: 8px solid #F4F4F4; border-right: 8px solid transparent; transform: translateY(16px);"></div> <div style=" background-color: #F4F4F4; flex-grow: 0; height: 12px; width: 16px; transform: translateY(-4px);"></div> <div style=" width: 0; height: 0; border-top: 8px solid #F4F4F4; border-left: 8px solid transparent; transform: translateY(-4px) translateX(8px);"></div></div></div> <div style="display: flex; flex-direction: column; flex-grow: 1; justify-content: center; margin-bottom: 24px;"> <div style=" background-color: #F4F4F4; border-radius: 4px; flex-grow: 0; height: 14px; margin-bottom: 6px; width: 224px;"></div> <div style=" background-color: #F4F4F4; border-radius: 4px; flex-grow: 0; height: 14px; width: 144px;"></div></div></a><p style=" color:#c9c8cd; font-family:Arial,sans-serif; font-size:14px; line-height:17px; margin-bottom:0; margin-top:8px; overflow:hidden; padding:8px 0 7px; text-align:center; text-overflow:ellipsis; white-space:nowrap;"><a href="{self.url}?utm_source=ig_embed&amp;utm_campaign=loading" style=" color:#c9c8cd; font-family:Arial,sans-serif; font-size:14px; font-style:normal; font-weight:normal; line-height:17px; text-decoration:none;" target="_blank">A post shared by {self.full_name} (@{self.username})</a></p></div></blockquote> <script async src="//www.instagram.com/embed.js"></script>'
        return html_template

    @staticmethod
    def _url_from_suburl(suburl: str) -> str:
        return f"https://www.instagram.com/p/{suburl}/"

    def _download_photo(self, fp: str, resp: requests.models.Response) -> None:
        with open(fp, "wb") as outfile:
            resp.raw.decode_content = True
            shutil.copyfileobj(resp.raw, outfile)

    def _download_video(self, fp: str, resp: requests.models.Response) -> None:
        """Write the media to file at given fp from the response"""
        with open(fp, "wb") as outfile:
            for chunk in resp.iter_content(chunk_size=1024):
                if chunk:
                    outfile.write(chunk)
                    outfile.flush()

    def _parse_tagged_users(self, json_dict: dict) -> List[str]:
        """Parse the tagged users from JSON dict containing the tagged users"""
        if "graphql" in json_dict:
            json_dict = [json_dict]
            json_dict = {"PostPage": json_dict}
            json_dict = {"entry_data": json_dict}
        tagged_arr = json_dict["entry_data"]["PostPage"][0]["graphql"]["shortcode_media"]["edge_media_to_tagged_user"][
            "edges"
        ]
        return [node["node"]["user"]["username"] for node in tagged_arr]

    def _parse_hashtags(self, caption: str) -> List[str]:
        """Parse the hastags from the post's caption using regex"""
        pattern = r"#(\w+)"
        return re.findall(pattern, caption)



from instascrape.scrapers.post import Post
from instascrape.core._mappings import _IGTVMapping


class IGTV(Post):
    """Scraper for an IGTV post"""

    _Mapping = _IGTVMapping

    @staticmethod
    def _url_from_suburl(suburl: str) -> str:
        return f"https://www.instagram.com/tv/{suburl}/"



"""
Algorithms and implementations for working with/manipulating JSON data. NOT
intended for top level use but instead imported for top level functions to leverage
"""

from collections import deque
from typing import Any, Dict, List, Union

from bs4 import BeautifulSoup

JSONDict = Dict[str, Any]

class _JSONTree:
    """Tree of linked lists that map out the JSON data"""

    def __init__(self, json_dict: JSONDict) -> None:
        self.json_dict = json_dict
        self.map_tree(self.json_dict)

    def map_tree(self, json_dict) -> None:
        """Map the entire JSON tree and get access to leaf _JSONNodes"""
        self.leaf_nodes = []
        self.root_node = _JSONNode(json_data=json_dict, tree=self)

class _JSONNode:
    """Representation of one step into a JSON Tree"""

    def __init__(self, json_data: Any, tree: _JSONTree, linked_list: deque = None, prior_keys: List[Union[str, int]] = None) -> None:
        self.json_data = json_data
        self.tree = tree

        self.linked_list = linked_list if linked_list is not None else deque([])
        self.prior_keys = prior_keys if prior_keys is not None else []

        self.dtype = type(self.json_data)

        self.nodes = []

        # If the node is a leaf then it has no edges
        if self.is_leaf:
            self.json_data = {prior_keys[-1]: self.json_data}
            self.tree.leaf_nodes.append(self)

        else:
            self.get_edges()

    @property
    def is_leaf(self) -> bool:
        """
        If the dtype of self.json_data is not a dict or a list then it must be
        a leaf node
        """
        return self.dtype is not list and self.dtype is not dict

    def get_edges(self) -> None:
        """Get all edges connected to current _JSONNode"""
        if self.dtype is list:
            iter_arr = zip(range(len(self.json_data)), self.json_data)
        else:
            iter_arr = self.json_data.items()

        for key, value in iter_arr:
            next_linked_list = self.linked_list + deque([self])
            next_key = self.prior_keys + [key]
            node = _JSONNode(value, self.tree, next_linked_list, next_key)
            self.nodes.append(node)

    def __repr__(self) -> str:
        return str(self.json_data)


def _parse_json_str(source: str) -> str:
    """Return the parsed string of JSON data from the BeautifulSoup"""
    json_data = []
    json_scripts = [str(script) for script in source.find_all("script") if "config" in str(script)]
    for script_tag in json_scripts:
        left_index = script_tag.find("{")
        right_index = script_tag.rfind("}") + 1
        json_str = script_tag[left_index:right_index]
        json_data.append(json_str)
    return json_data



from __future__ import annotations

import datetime
import json
import csv
from abc import ABC, abstractmethod
from typing import Union, Dict, List, Any
import sys
import os
from collections import namedtuple
import warnings

import requests
from bs4 import BeautifulSoup

from instascrape.scrapers.scrape_tools import parse_data_from_json, determine_json_type, flatten_dict, json_from_soup
from instascrape.exceptions.exceptions import InstagramLoginRedirectError, MissingSessionIDWarning, MissingCookiesWarning

# pylint: disable=no-member

JSONDict = Dict[str, Any]

class _StaticHtmlScraper(ABC):
    """
    Base class for all of the scrapers, handles general functionality that all
    scraper objects will have
    """

    # Keys that represent metadata attr that the user doesn't necessarily need
    # to worry about
    _METADATA_KEYS = [
        "json_dict",
        "url",
        "_json_scraper",
        "scrape_timestamp",
        "map_dict",
        "json_data",
        "json_flattener",
        "flat_json_dict",
        "soup",
        "html",
        "source",
    ]
    _ASSOCIATED_JSON_TYPE = None

    session = requests.Session()

    def __init__(self, source: Union[str, BeautifulSoup, JSONDict]) -> None:
        """
        Parameters
        ----------
        source : Union[str, BeautifulSoup, JSONDict]
            The given source for scraping the data from. Available sources are
            a URL, HTML, JSON dictionary, BeautifulSoup, etc.
        """
        self.source = source

        # Instance variables that are given values elsewhere
        self.url = None
        self.html = None
        self.soup = None
        self.json_dict = None
        self.flat_json_dict = None
        self.scrape_timestamp = None

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)

    def __repr__(self) -> str:
        return f"<{type(self).__name__}>"

    def scrape(
        self,
        mapping=None,
        keys: List[str] = None,
        exclude: List[str] = None,
        headers={
            "user-agent": "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Mobile Safari/537.36 Edg/87.0.664.57"
        },
        inplace=True,
        session=None,
        webdriver=None
    ) -> None:
        """
        Scrape data from the source

        Parameters
        ----------
        mapping : Dict[str, deque]
            Dictionary of parsing queue's that tell the JSON engine how to
            process the JSON data
        keys : List[str]
            List of strings that correspond to desired attributes for scraping
        exclude : List[str]
            List of strings that correspond to which attributes to exclude from
            being scraped
        headers : Dict[str, str]
            Dictionary of request headers to be passed on the GET request
        inplace : bool
            Determines if data modified inplace or return a new object with the
            scraped data
        session : requests.Session
            Session for making the GET request
        webdriver : selenium.webdriver.chrome.webdriver.WebDriver
            Webdriver for scraping the page, overrides any default or passed
            session

        Returns
        -------
        return_instance
            Optionally returns a scraped instance instead of modifying inplace
            if inplace arg is True
        """

        if mapping is None:
            mapping = self._Mapping.return_mapping(keys=keys, exclude=exclude)
        if session is None:
            session = self.session
        if webdriver is not None:
            session = webdriver
        if keys is None:
            keys = []
        if exclude is None:
            exclude = []

        if webdriver is None:
            try:
                if "sessionid" not in headers["cookie"]:
                    warnings.warn(
                        "Session ID not in cookies! It's recommended you pass a valid sessionid otherwise Instagram will likely redirect you to their login page.",
                        MissingSessionIDWarning
                    )
            except KeyError:
                warnings.warn(
                    "Request header does not contain cookies! It's recommended you pass at least a valid sessionid otherwise Instagram will likely redirect you to their login page.",
                    MissingCookiesWarning
                    )

        # If the passed source was already an object, construct data from
        # source else parse it
        if isinstance(self.source, type(self)):
            scraped_dict = self.source.to_dict()
        else:
            return_data = self._get_json_from_source(self.source, headers=headers, session=session)
            flat_json_dict = flatten_dict(return_data["json_dict"])
            scraped_dict = parse_data_from_json(
                json_dict=flat_json_dict,
                map_dict=mapping,
            )
        return_data["scrape_timestamp"] = datetime.datetime.now()
        return_data["flat_json_dict"] = flat_json_dict
        return_instance = self._load_into_namespace(
                            scraped_dict=scraped_dict,
                            return_data=return_data,
                            inplace=inplace
        )
        return None if return_instance is self else return_instance

    def to_dict(self, metadata: bool = False) -> Dict[str, Any]:
        """
        Return a dictionary containing all of the data that has been scraped

        Parameters
        ----------
        metadata : bool
            Boolean value that determines if metadata specified in self._METADATA_KEYS
            will be included in the dictionary.

        Returns
        -------
        data_dict : Dict[str, Any]
            Dictionary containing the scraped data
        """
        data_dict = (
            {key: val for key, val in self.__dict__.items() if key not in self._METADATA_KEYS}
            if not metadata
            else self.__dict__
        )
        return data_dict

    def to_csv(self, fp: str) -> None:
        """
        Write scraped data to .csv at the given filepath

        Parameters
        ----------
        fp : str
            Filepath to write data to
        """
        with open(fp, "w", newline="", encoding="utf-8") as csv_file:
            writer = csv.writer(csv_file)
            for key, value in self.to_dict().items():
                writer.writerow([key, str(value)])

    def to_json(self, fp: str) -> None:
        """
        Write scraped data to .json file at the given filepath

        Parameters
        ----------
        fp : str
            Filepath to write data to
        """
        outdict = {key: str(val) for key, val in self.to_dict().items()}
        with open(fp, "w") as outjson:
            json.dump(outdict, outjson)

    @abstractmethod
    def _url_from_suburl(self, suburl: str) -> str:
        pass

    def _get_json_from_source(self, source: Any, headers: dict, session: requests.Session) -> JSONDict:
        """Parses the JSON data out from the source based on what type the source is"""
        initial_type = True
        return_data = {"source": self.source}
        if isinstance(source, str):
            source_type = self._determine_string_type(source)
        elif isinstance(source, dict):
            json_dict = source
            source_type = "json dict"
        elif isinstance(source, BeautifulSoup):
            source_type = "soup"

        if source_type == "suburl":
            if initial_type:
                suburl = self.source
            url = self._url_from_suburl(suburl=suburl)
            source_type = "url"
            initial_type = False
            return_data["url"] = url

        if source_type == "url":
            if initial_type:
                url = self.source
            html = self._html_from_url(url=url, headers=headers, session=session)
            source_type = "html"
            initial_type = False
            return_data["html"] = html

        if source_type == "html":
            if initial_type:
                html = self.source
            soup = self._soup_from_html(html)
            source_type = "soup"
            initial_type = False
            return_data["soup"] = soup

        if source_type == "soup":
            if initial_type:
                soup = self.source
            json_dict_arr = json_from_soup(soup)
            if len(json_dict_arr) == 1:
                json_dict = json_dict_arr[0]
            else:
                json_dict = json_dict_arr[1]
            self._validate_scrape(json_dict)

        return_data["json_dict"] = json_dict

        return return_data

    def _load_into_namespace(self, scraped_dict: dict, return_data, inplace) -> None:
        """Loop through the scraped dictionary and set them as instance attr"""
        instance = self if inplace else type(self)(return_data["source"])
        for key, val in scraped_dict.items():
            setattr(instance, key, val)
        for key, val in return_data.items():
            setattr(instance, key, val)
        return instance


    @staticmethod
    def _html_from_url(url: str, headers: dict, session: requests.Session) -> str:
        """Return HTML from requested URL"""
        if isinstance(session, requests.Session):
            response = session.get(url, headers=headers)
            page_source = response.text
        else:
            session.get(url)
            page_source = session.page_source
        return page_source

    @staticmethod
    def _soup_from_html(html: str) -> BeautifulSoup:
        """Return BeautifulSoup from source HTML"""
        return BeautifulSoup(html, features="html.parser")

    def _validate_scrape(self, json_dict: str) -> JSONDict:
        """Raise exceptions if the scrape did not properly execute"""
        json_type = determine_json_type(json_dict)
        if json_type == "LoginAndSignupPage" and not type(self).__name__ == "LoginAndSignupPage":
            raise InstagramLoginRedirectError
        elif json_type == "HttpErrorPage" and not type(self).__name__ == "HttpErrorPage":
            source_str = self.url if hasattr(self, "url") else "Source"
            raise ValueError(f"{source_str} is not a valid Instagram page. Please provide a valid argument.")

    @staticmethod
    def _determine_string_type(string_data: str) -> str:
        """Match and return string representation of appropriate source"""
        string_type_map = [("https://", "url"), ("window._sharedData", "html"), ('{"config"', "JSON dict str")]
        for substr, str_type in string_type_map:
            if substr in string_data:
                #BUG: !DOCTYPE isnt returned in selenium source code, use </html> as secondary identifier instead
                if substr == "https://" and "!DOCTYPE" in string_data:
                    continue
                break
        else:
            str_type = "suburl"
        return str_type



from __future__ import annotations

import json
from typing import Any, Dict, Union
from collections import deque

import requests
from bs4 import BeautifulSoup

from instascrape.core.json_algos import _JSONTree, _parse_json_str

JSONDict = Dict[str, Any]

def parse_data_from_json(json_dict, map_dict, default_value=float('nan')):
    """
    Parse data from a JSON dictionary using a mapping dictionary that tells
    the program how to parse the data
    """
    return_data = {}
    for key in map_dict:
        steps_to_value = map_dict[key]

        # Loop through all steps into the JSON dict that will give us our data
        first_step = steps_to_value.popleft()
        try:
            value = json_dict[first_step]
        except KeyError:
            value = default_value
        else:
            for step in steps_to_value:
                value = json_dict[step]
        finally:
            return_data[key] = value
    return return_data

def flatten_dict(json_dict: JSONDict) -> JSONDict:
    """
    Returns a flattened dictionary of data

    Parameters
    ----------
    json_dict : dict
        Input dictionary for flattening

    Returns
    -------
    flattened_dict : dict
        Flattened dictionary
    """
    json_tree = _JSONTree(json_dict)
    flattened_dict = {}
    for leaf_node in json_tree.leaf_nodes:
        key_arr = deque([])
        for key in leaf_node.prior_keys[::-1]:
            key_arr.appendleft(str(key))
            new_key = "_".join(key_arr)
            if new_key not in flattened_dict:
                break
        flattened_dict[new_key] = list(leaf_node.json_data.values())[0]
    return flattened_dict

def json_from_html(source: Union[str, "BeautifulSoup"], as_dict: bool = True, flatten=False) -> Union[JSONDict, str]:
    """
    Return JSON data parsed from Instagram source HTML

    Parameters
    ----------
    source : Union[str, BeautifulSoup]
        Instagram HTML source code to parse the JSON from
    as_dict : bool = True
        Return JSON as dict if True else return JSON as string
    flatten : bool
        Flatten the dictionary prior to returning it

    Returns
    -------
    json_data : Union[JSONDict, str]
        Parsed JSON data from the HTML source as either a JSON-like dictionary
        or just the string serialization
    """

    soup = BeautifulSoup(source, features="html.parser")
    json_data = json_from_soup(source=soup, as_dict=as_dict, flatten=flatten)
    return json_data

def json_from_soup(source, as_dict: bool = True, flatten=False):
    json_data = _parse_json_str(source=source)

    if as_dict:
        json_data = [json.loads(json_str) for json_str in json_data]
    if flatten:
        json_data = [flatten_dict(json_dict) for json_dict in json_data]

    return json_data

def determine_json_type(json_data: Union[JSONDict, str]) -> str:
    """
    Return the type of Instagram page based on the JSON data parsed from source

    Parameters
    ----------
    json_data: Union[JSONDict, str]
        JSON data that will be checked and parsed to determine what type of page
        the program is looking at (Profile, Post, Hashtag, etc)

    Returns
    -------
    instagram_type : str
        Name of the type of page the program is currently parsing or looking at
    """
    if not isinstance(json_data, dict):
        json_data = json.loads(json_data)
    try:
        instagram_type = list(json_data["entry_data"])[0]
    except KeyError:
        instagram_type = "Inconclusive"
    return instagram_type

def json_from_url(
    url: str,
    as_dict: bool = True,
    headers={
        "user-agent": "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88 Mobile Safari/537.36 Edg/87.0.664.57"
    },
    flatten=False
) -> Union[JSONDict, str]:
    """
    Return JSON data parsed from a provided Instagram URL

    Parameters
    ----------
    url : str
        URL of the page to get the JSON data from
    as_dict : bool = True
        Return JSON as dict if True else return JSON as string
    headers : Dict[str, str]
            Dictionary of request headers to be passed on the GET request
    flatten : bool
        Flatten the dictionary prior to returning it

    Returns
    -------
    json_data : Union[JSONDict, str]
        Parsed JSON data from the URL as either a JSON-like dictionary
        or just the string serialization
    """
    source = requests.get(url, headers=headers).text
    return json_from_html(source, as_dict=as_dict, flatten=flatten)







"""
Parse data related to comments, including comments in a thread
"""

import datetime


class Comment:
    """A single comment and its respective data"""

    # pylint: disable=too-many-instance-attributes, too-few-public-methods

    def __init__(self, comment_dict: dict) -> None:
        self.comment_dict = comment_dict["node"]

        self._parse_data()

    def __repr__(self) -> str:
        return f"<Comment: {self.username}: {self.text}"

    def _parse_data(self) -> None:
        self.text = self.comment_dict["text"]
        self.created_at = datetime.datetime.fromtimestamp(self.comment_dict["created_at"])
        self.did_report_as_spam = self.comment_dict["did_report_as_spam"]
        self.is_verified = self.comment_dict["owner"]["is_verified"]
        self.profile_pic_url = self.comment_dict["owner"]["profile_pic_url"]
        self.username = self.comment_dict["owner"]["username"]
        self.viewer_has_liked = self.comment_dict["viewer_has_liked"]
        self.likes = self.comment_dict["edge_liked_by"]["count"]
        self.is_restricted_pending = self.comment_dict["is_restricted_pending"]

        try:
            comments = self.comment_dict["edge_threaded_comments"]["edges"]
            self.replies = [Comment(comment_dict) for comment_dict in comments]
        except KeyError:
            self.replies = []



class InstagramLoginRedirectError(Exception):
    """
    Exception that indicates Instagram is redirecting away from the page
    that should be getting scraped. Can be remedied by logging into Instagram.
    """

    def __init__(
        self,
        message="Instagram is redirecting you to the login page instead of the page you are trying to scrape. This could be occuring because you made too many requests too quickly or are not logged into Instagram on your machine. Try passing a valid session ID to the scrape method as a cookie to bypass the login requirement",
    ):
        super().__init__(message)


class WrongSourceError(Exception):
    """
    Exception that indicates user passed the wrong source type to the scraper.
    An example is passing a URL for a hashtag page to a Profile.
    """

    def __init__(self, message="Wrong input source, use the correct class"):
        super().__init__(message)


class MissingSessionIDWarning(UserWarning):
    pass

class MissingCookiesWarning(UserWarning):
    pass


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="insta-scrape",
    version="2.0.0",
    author="Chris Greening",
    author_email="chris@christophergreening.com",
    description="Super lightweight Instagram web scraper for data analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/chris-greening/instascrape",
    packages=["instascrape", "instascrape.core", "instascrape.scrapers", "instascrape.exceptions"],
    install_requires=["requests", "beautifulsoup4"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)



"""
Primary API scraper tools
"""

from instascrape.scrapers.hashtag import *
from instascrape.scrapers.post import *
from instascrape.scrapers.profile import *
from instascrape.scrapers.comment import *
from instascrape.scrapers.location import *
from instascrape.scrapers.reel import *
from instascrape.scrapers.igtv import *
from instascrape.scrapers.scrape_tools import *